#!/bin/bash

#  Testing script for the Lambert W function
#
#  Copyright (C) 2012 Darko Veberic, darko.veberic@ung.si
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


# This bash script reads-in two reference data files, one for each branch
# (0 and -1) of the Lambert W function. Files contain lines with values
# for x and W(x). These values are then within an awk script compared to
# the return values of the lambert-w command-line utility and a non-zero
# return value is issued when discrepancies larger than accuracy are
# encountered. Accuracy is the only option ($1) to this script.

if [ -z "$1" ] ; then
  echo "No accuracy specified!"
  exit 1
fi

accuracy="$1"

for branch in 0 m ; do
  case $branch in
  (0) b=0 ;;
  (m) b="-1" ;;
  esac
  echo -n "branch $b: "
  awk '
BEGIN {
  eps = '$accuracy'
}
function abs(x) {
  return x < 0 ? -x : x
}
{
  c = "./lambert-w '$b' "$1
  c | getline w
  if (w != w+0) {
    print "Result not a number: lambert-w '$b' "$1" = "w
    failed = 1
    exit(2)
  }
  d = (abs(w) < eps) ? abs(w - $2) : abs((w - $2) / $2)
  if (d > eps) {
    print "Difference test failed: lambert-w '$b' "$1" = "w", but "$2" was expected!"
    failed = 1
    exit(1)
  }
}
END {
  if (!failed)
    print "ok"
}
' testData_$branch.dat || exit 2
done
